#include "env.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pebblisp.h"
#include "web.h"
#include "plfunc.h"

struct Environment* globalEnv;

struct Environment* global()
{
    return globalEnv;
}

void setGlobal(struct Environment* env)
{
    globalEnv = env;
}

struct Dictionary dictionary;

Object fetchFromEnvironment(const char* name, struct Environment* env)
{
    if (!env) {
        return errorObject(NULL_ENV);
    }

    if (env->capacity == 0) {
        if (env->outer) {
            return fetchFromEnvironment(name, env->outer);
        } else {
            return errorObject(EMPTY_ENV);
        }
    }

    for (int i = 0; i < env->capacity; i++) {
        if (env->strings[i] == NULL) { printd("Try %d (NULL)\n", i);
            break;
        }

        printd("Try %d (%s)\n", i, env->strings[i]);
        debugObj(&env->objects[i]);
        if (strcmp(name, env->strings[i]) == 0) { printd("Returning!\n");
            return cloneObject(env->objects[i]);
        }
    }

    printd("Trying outer %p\n", env->outer);
    if (env->outer) {
        return fetchFromEnvironment(name, env->outer);
    }

    return errorWithContext(DID_NOT_FIND_SYMBOL, name);
}

struct Environment envForLambda(const Object* params, const Object* arg_forms,
                                struct Environment* outer)
{
    int paramCount = listLength(params);

    if (outer) {
        outer->refs += 1;
    }

    struct Environment env = {
        .outer = outer,
        .capacity = paramCount,
        .refs = 1,
    };

    if (paramCount == 0) {
        if (outer) {
            return *outer;
        }
        env.strings = NULL;
        env.objects = NULL;
        return env;
    }

    env.strings = calloc(sizeof(char*), paramCount + 1);
    env.objects = malloc(sizeof(Object) * (paramCount + 1));

    const Object* march = arg_forms;
    for (int i = 0; i < paramCount; i++) {
        const char* newObjName = itemAt(params, i)->string;
        // Eval the `march` list
        Object newEnvObj = march ? eval(march, outer) : errorObject(NOT_ENOUGH_ARGUMENTS);
        addToEnv(&env, newObjName, newEnvObj); // Could use eval_forms?
        cleanObject(&newEnvObj);
        march = march ? march->forward : NULL;
    }

    return env;
}

void addToEnvAt(int i, struct Environment* env, const char* name, const Object obj)
{
    env->strings[i] = malloc(sizeof(char) * strlen(name) + 1);
    strcpy(env->strings[i], name);
    env->objects[i] = cloneObject(obj);
}

void addToEnv(struct Environment* env, const char* name, const Object obj)
{
    struct Environment* temp_env = env;
    while (temp_env) {
        for (int i = 0; i < temp_env->capacity; i++) {
            if (temp_env->strings[i] == NULL) {
                // Add *new* item to env only if we're in the original scope
                if (temp_env == env) {
                    addToEnvAt(i, env, name, obj);
                    return;
                }
                break;
            } else if (strcmp(temp_env->strings[i], name) == 0) {
                Object o = cloneObject(obj);
                cleanObject(&temp_env->objects[i]);
                temp_env->objects[i] = o;
                return;
            }
        }
        temp_env = temp_env->outer;
    }

    printd("Reallocating environment\n");
    const int inc = 5;
    const int old_size = env->capacity;
    env->capacity += inc;
    env->strings = realloc(env->strings, sizeof(char*) * env->capacity);
    env->objects = realloc(env->objects, sizeof(Object) * env->capacity);

    for (int j = 0; j < inc; j++) {
        env->strings[old_size + j] = NULL;
    }

    addToEnvAt(old_size, env, name, obj);
}

void printEnv(struct Environment* env)
{
    if (!env) {
        printf("NULL env\n");
        return;
    }
    printf("env->capacity = %d\n", env->capacity);
    for (int i = 0; i < env->capacity; i++) {
        if (env->strings[i] == NULL) {
            printf("env[%d]: NULL %p\n", i, env->strings[i]);
            break;
        }
        printf("env[%d]: `%s` %p :: ", i, env->strings[i], env->strings[i]);
        printf("[1m");
        printObj(&env->objects[i]);
        printf("[0m");
    }
}

void addFunc(const char* name,
             Object (* func)(Object*, int, struct Environment*),
             struct Environment* env,
             int i)
{
    Object o = newObject(TYPE_FUNC);
    o.func = func;
    addToEnvAt(i, env, name, o);
}

void deleteEnv(struct Environment* e)
{
    e->refs -= 1;
    if (e->refs) {
        return;
    }
    if (e->outer) {
        deleteEnv(e->outer);
    }

    if (e->strings) {
        int i = 0;
        while (e->strings[i]) {
            free(e->strings[i]);
            cleanObject(&e->objects[i]);
            i++;
        }
        free(e->strings);
        free(e->objects);
    }
}

void shredDictionary()
{
    for (int i = 0; i < dictionary.structCount; i++) {
        free(dictionary.structDefs[i].name);
        for (int j = 0; j < dictionary.structDefs[i].fieldCount; j++) {
            free(dictionary.structDefs[i].names[j]);
        }
        free(dictionary.structDefs[i].names);
    }
    free(dictionary.structDefs);
}

struct symFunc {
    const char* sym;

    Object (* func)(Object*, int, struct Environment*);
};

struct helpText {
    const char* symbol;
    const char* help;
    const char* const* tests;
    size_t testCount;
};
int currentHelp = 0;
struct helpText helpTexts[100];

#define pfn(_func) buildFuncSym(_func ## Symbol, &(_func), _func ## Doc, NULL)
#define pf(_name, _func) buildFuncSym(_name, &(_func), _func ## Doc, _func ## Tests, array_length(_func ## Tests))

/// For any instances (e.g. segfault recovery) where defaultEnv() may be called
/// multiple times.
int helpInitialized = 0;

struct symFunc buildFuncSym(const char* symbol, Object (* func)(Object*, int, struct Environment*), const char* help,
                            const char* const* tests, size_t testLength)
{
    if (!helpInitialized) {
        helpTexts[currentHelp].help = help;
        helpTexts[currentHelp].symbol = symbol;
        helpTexts[currentHelp].tests = tests;
        helpTexts[currentHelp].testCount = testLength;
        currentHelp += 1;
    }

    return (struct symFunc) {
        .func = func,
        .sym = symbol,
    };
}

const int black = 30;
const int red = 31;
const int green = 32;
const int yellow = 33;
const int cyan = 34;
const int purple = 35;
const int teal = 36;
const int white = 37;
const int colors[] = {
    white, green, cyan, yellow, purple, red
};
const int colorCount = array_length(colors);

int getColor(int depth)
{
    depth = depth >= 0 ? depth : depth * -1;
    return colors[depth % colorCount];
}

void printColored(const char* code)
{
    int c = 0;
    int depth = 0;
    printf("[%dm", getColor(depth));
    while (code[c]) {
        if (code[c] == '(') {
            depth += 1;
            printf("[%dm", getColor(depth));
        } else if (code[c] == ')') {
            depth -= 1;
            printf(")[%dm", getColor(depth));
            c++;
            continue;
        } else if (code[c] == '=' && code[c + 1] == '>') {
            printf("[1m=>[0;%dm", getColor(depth));
            c += 2;
            continue;
        }
        printf("%c", code[c]);
        c++;
    }
}

char* getHelp(const char* symbol)
{
    for (int i = 0; i < currentHelp; i++) {
        struct helpText h = helpTexts[i];
        if (strcmp(symbol, h.symbol) == 0) {
            char* text = calloc(sizeof(char), 1024);
            char* textCursor = text;
            textCursor += sprintf(textCursor, "%s", h.help);
            for (int ti = 0; ti < h.testCount; ti += 2) {
                const char* test = h.tests[ti];
                const char* expected = h.tests[ti + 1];
                textCursor += sprintf(textCursor, "\n  %s => ", test);
                if (expected[0] == '\0') {
                    textCursor += sprintf(textCursor, "<empty string>");
                } else {
                    textCursor += sprintf(textCursor, "%s", expected);
                }
            }
            return text;
        }
    }
    const char* notFound = "Help not found!";
    char* help = malloc(sizeof(char) * (strlen(notFound) + 1));
    sprintf(help, "%s", notFound);
    return help;
}

fnn(segfault, "seg", "Induces a segfault.")
{
    int* p = NULL;
    return numberObject(*p);
}

// Returns number of failures
int runTests(int detailed)
{
    printf("[35;1m::NATIVE TESTS::[0m\n");
    int failureCount = 0;
    int passCount = 0;
    for (int hi = 0; hi < currentHelp; hi++) {
        struct helpText h = helpTexts[hi];
        if (h.tests) {
            char result[1024];
            if (detailed && h.testCount > 0) {
                printf("[0m `%s` ", h.symbol);
            }
            for (int ti = 0; ti < h.testCount; ti += 2) {
                const char* test = h.tests[ti];
                const char* expected = h.tests[ti + 1];
                struct Environment env = defaultEnv();
                Object o = parseEval(test, &env);
                stringObj(result, &o);
                cleanObject(&o);
                if (strcmp(result, expected) != 0) {
                    failureCount++;
                    printf("Test failed!\n");
                    printf("%s\n", test);
                    printf("Expected '%s' but received '%s'\n", expected, result);
                } else {
                    if (detailed) {
                        printf("[32;1m✓");
                    }
                    passCount++;
                }
                deleteEnv(&env);
            }
            if (detailed && h.testCount > 0) {
                printf("\n");
            }
        }
    }
    if (passCount > 0) {
        printf("[32;1m");
    }
    printf("%d tests passed![0m\n", passCount);
    if (failureCount > 0) {
        printf("[31m");
    }
    printf("%d tests failed![0m\n", failureCount);
    // fprintf(stderr, "TOTAL ALLOCATIONS: %d\n", getAllocations());
    // fprintf(stderr, "TOTAL BYTES: %zu\n", getBytes());
    return failureCount;
}

struct Environment defaultEnv()
{
    if (!helpInitialized) {
        dictionary = (struct Dictionary) {
            .structCount = 0,
            .structCapacity = 8,
            .structDefs = malloc(sizeof(struct StructDef) * 8),
        };
    }

    struct Environment e = {
        .outer = NULL,
        .strings = calloc(sizeof(char*), MAX_ENV_ELM),
        .objects = malloc(sizeof(Object) * MAX_ENV_ELM),
        .capacity = MAX_ENV_ELM,
        .refs = 1,
    };

    struct symFunc symFuncs[] = {
        {"+", &add},
        {"-", &sub},
        {"*", &mul},
        {"/", &dvi},
        {"%", &mod},
        {"=", &equ},
        {">", &greaterThan},
        {"<", &lessThan},
        {"&", &and},
        {"|", &or},
        pf("cat", catObjects),
        pf("fil", filter),
        pf("len", len),
        pf("ap", append),
        pf("pre", prepend),
        pf("reduce", reduce),
        pf("at", at),
        pf("rest", rest),
        pf("chat", charAt),
#ifndef LOW_MEM
        pf("rev", reverse),
#endif
        pf("isnum", isNum),
        pf("islist", isList),
        pf("isstr", isString),
        pf("iserr", isErr),
        pf("char", charVal),
        pf("eval", parseEvalO),
        pf("poss", possessive),
#ifdef WEBSERVER
        pf("get", addGetRoute),
        pf("post", addPostRoute),
        pf("serve", startServer),
#endif
#ifdef STANDALONE
        pf("seg", segfault),
        //pfn(segfault),
        pf("prn", print),
        pf("ch", numToChar),
        pf("penv", printEnvO),
        pf("sys", systemCall),
        pf("loadfile", loadFile),
        pf("inp", takeInput),
        pf("rf", readFileToObject),
        pf("?", help)
#endif
    };

    for (int i = 0; i < sizeof(symFuncs) / sizeof(symFuncs[0]); i++) {
        addFunc(symFuncs[i].sym, symFuncs[i].func, &e, i);
    }

    helpInitialized = 1;
    return e;
}

int getStructIndex(const char* name)
{
    for (int i = 0; i < dictionary.structCount; i++) {
        if (strcmp(name, dictionary.structDefs[i].name) == 0) {
            return i;
        }
    }
    return -1;
}

struct StructDef* getStructAt(int i)
{
    return &dictionary.structDefs[i];
}

void addStructDef(struct StructDef def)
{
    dictionary.structDefs[dictionary.structCount] = def;
    dictionary.structCount += 1;
    if (dictionary.structCount == dictionary.structCapacity) {
        struct StructDef* prev = dictionary.structDefs;
        int prevCapacity = dictionary.structCapacity;
        dictionary.structCapacity *= 2;
        dictionary.structDefs = malloc(sizeof(struct StructDef) * dictionary.structCapacity);
        for (int i = 0; i < prevCapacity; i++) {
            dictionary.structDefs[i] = prev[i];
        }
        free(prev);
    }

}
