#include "object.h"
#include "pebblisp.h"
#include "env.h"

#include <stdio.h>
#include <string.h>

size_t bytes = 0;

size_t getBytes()
{
    return bytes;
}

void* smalloc(size_t size)
{
    bytes += size;
    return malloc(size);
}

void* scalloc(size_t size, size_t count)
{
    bytes += (size * count);
    return calloc(size, count);

}

#define malloc(x) smalloc(x)
#define calloc(x, y) scalloc(x, y)

/**
 * Returns the length of a given list Object
 * @param listObj The list to get the length of
 * @return Length of the list if non-null and of the list type. Otherwise -1
 */
int listLength(const Object* listObj)
{
    int len = 0;
    FOR_POINTER_IN_LIST(listObj) {
        len++;
    }
    return len;
}

/**
 * Returns a pointer to the Object at the given index in a given list
 * @param listObj The list to fetch an Object from
 * @param n The index to to fetch from the list
 * @return A pointer to the Object, if it is found, or NULL on an error
 */
Object* itemAt(const Object* listObj, int n)
{
    if (!listObj || !isListy(*listObj)) {
        return NULL;
    }

    FOR_POINTER_IN_LIST(listObj) {
        if (n-- == 0) {
            return POINTER;
        }
    }
    return NULL;
}

/**
 * Returns a pointer to the last Object in the given list.
 * @param listObj The list to find the tail of
 * @return A pointer to the last Object if it is found, or NULL on an error
 */
Object* tail(const Object* listObj)
{
    if (!listObj || !isListy(*listObj)) {
        return NULL;
    }

    Object* tail = NULL;
    FOR_POINTER_IN_LIST(listObj) {
        tail = POINTER;
    }
    return tail;
}

/**
 * Checks if a given object is empty.
 *
 * BOOL, NUMBER, and ERROR types return true only when 0
 * LIST, LAMBDA, and FUNC types return true only when NULL
 * SYMBOL returns true only when it begins with a nullchar
 * @param obj A pointer to the object to check
 * @return True only if the non-null Object is empty/non-zero
 */
inline int isEmpty(const Object* obj)
{
    if (obj == NULL) {
        return 0;
    }

    switch (obj->type) {
        case TYPE_NUMBER:
        case TYPE_BOOL:
            return obj->number == 0;
        case TYPE_LIST:
        case TYPE_SLIST:
            return obj->list == NULL;
        case TYPE_LAMBDA:
            return obj->lambda == NULL;
        case TYPE_STRUCT:
            return obj->structObject == NULL;
        case TYPE_SYMBOL:
        case TYPE_STRING:
            return obj->string == NULL || obj->string[0] == '\0';
        case TYPE_FUNC:
            return obj->func == NULL;
        case TYPE_OTHER:
            return obj->other == NULL;
        case TYPE_ERROR:
            return getErrorCode(*obj);
    }

    return 0;
}

int allocations = 0;

int getAllocations()
{
    return allocations;
}

/**
 * Allocate a copy of a given object into the given pointer.
 * Does nothing if `spot` is NULL
 * @param spot A pointer to the Object pointer that needs allocating
 * @param src The Object to copy from
 */
void allocObject(Object** spot, const Object src)
{
#ifdef ALLOCATION_CAP
    if (allocations >= 10000) {
        printf("MAX ALLOCATIONS EXCEEDED\n");
        *spot = NULL;
        return;
    }
#endif
    allocations++;
    *spot = malloc(sizeof(struct Object));
    **spot = src;
    (*spot)->forward = NULL;
}

/**
 * Adds an Object to the end of a list
 * Does nothing if `dest` is NULL or not a list type
 * @param dest The list to append to
 * @param src The Object to copy into the list
 */
void nf_addToList(Object* dest, const Object src)
{
    if (!dest || !isListy(*dest)) {
        return;
    }

    if (isEmpty(dest)) {
        allocObject(&dest->list, src);
        return;
    }

    allocObject(&tail(dest)->forward, src);
}

#ifndef SIMPLE_ERRORS
static const char* errorText[] = {"MISMATCHED_PARENS",
                                  "NULL_ENV",
                                  "EMPTY_ENV",
                                  "NULL_PARSE",
                                  "NULL_LAMBDA_LIST",
                                  "NULL_MAP_ARGS",
                                  "LAMBDA_ARGS_NOT_LIST",
                                  "DID_NOT_FIND_SYMBOL",
                                  "BAD_TYPE",
                                  "BAD_PARAMS_ON",
                                  "BAD_NUMBER",
                                  "UNSUPPORTED_NUMBER_TYPE",
                                  "NOT_ENOUGH_ARGUMENTS",
                                  "NOT_A_LIST",
                                  "SCRIPT_NOT_FOUND",
                                  "NO_CLONE_SPECIFIED",
                                  "CAN_ONLY_EVAL_STRINGS",
                                  "UNEXPECTED_EOF",
                                  "INDEX_PAST_END"};
#endif

/**
 * Creates a string from a given list Object
 * Blocks out with parens
 * Returns immediately if `dest` is NULL, or `obj` is not a list type
 * @param dest The string to copy the list text into
 * @param obj The list Object to make a string from
 */
int stringList(char* dest, const Object* obj)
{
    if (!dest || !isListy(*obj)) {
        return 0;
    }
    const char* initial = dest;

    dest += sprintf(dest, "(");

    FOR_POINTER_IN_LIST(obj) {
        char tok[256] = "";
        stringObj(tok, POINTER);
        dest += sprintf(dest, " %s", tok);
    }
    dest += sprintf(dest, " )");
    return dest - initial;
}

/**
 * Creates a string from a given struct Object
 * Blocks out with curly braces
 */
int stringStruct(char* dest, const Object* obj)
{
    struct StructObject* so = obj->structObject;
    const char* initial = dest;
    dest += sprintf(dest, "{");

    struct StructDef* def = getStructAt(so->definition);
    for (int i = 0; i < def->fieldCount; i++) {
        dest += sprintf(dest, " %s: ", def->names[i]);
        int isString = so->fields[i].type == TYPE_STRING;
        if (isString) {
            dest += sprintf(dest, "\"");
        }

        char tok[90] = "";
        stringObj(tok, &so->fields[i]);
        dest += sprintf(dest, "%s", tok);

        if (isString) {
            dest += sprintf(dest, "\"");
        }
        dest += sprintf(dest, ",");
    }
    dest--;
    dest += sprintf(dest, " }");
    return dest - initial;
}
/**
 * Creates a string from a given Object
 * Returns NULL if either param is NULL
 *
 * Prints
 * -# Numbers without spaces
 * -# Symbols as their name
 * -# Bools as 'T' or 'F'
 * -# Strings as their string value
 * -# Lists using stringList()
 * -# Errors as the integer value, prepended with 'E'
 * -# Otherwise as the raw number, prepended with 'X'
 *
 * @param dest The string to copy the list text into
 * @param obj The list Object to make a string from
 */
#ifdef STANDALONE
#define stringf(_dest, _len, _format, args...) sprintf(_dest, _format, ## args)
#else
#define stringf(_dest, _len, _format, args...) snprintf(_dest, _len, _format, ## args)
#endif

int stringNObj(char* dest, const Object* obj, const size_t len)
{
    const char* initial = dest;

    switch (obj->type) {
        case TYPE_NUMBER:
            dest += stringf(dest, len, "%d", obj->number);
            break;
        case TYPE_BOOL:
            dest += stringf(dest, len, "%s", obj->number ? "T" : "F");
            break;
        case TYPE_STRING:
            dest += stringf(dest, len, "%s", obj->string);
            break;
        case TYPE_SYMBOL:
            dest += stringf(dest, len, "`%s`", obj->string);
            break;
        case TYPE_STRUCT:
            //snprintf(dest, len, "{%s}", obj->structObject->definition->names[0]);
            dest += stringStruct(dest, obj);
            break;
        case TYPE_LIST:
        case TYPE_SLIST:
            dest += stringList(dest, obj);
            break;
        case TYPE_ERROR: {
            int code = getErrorCode(*obj);
#ifdef SIMPLE_ERRORS
            dest += stringf(dest, len, "E[%d]", (int)code);
#else
            if (obj->error->context && obj->error->context[0] != '\0') {
                dest += stringf(dest, len, "%s: %s", errorText[code],
                                obj->error->context);
            } else if (code >= 0 && code <= INDEX_PAST_END) {
                dest += stringf(dest, len, "%s", errorText[code]);
            } else {
                dest += stringf(dest, len, "BROKEN ERROR CODE: %d", code);
            }
#endif
            break;
        }
        case TYPE_FUNC:
            dest += stringf(dest, len, "F%d", obj->number);
            break;
        case TYPE_LAMBDA: {
#ifdef STANDALONE
#ifdef DEBUG
            dest += stringf(dest, len, "\\x%d", obj->number);
#endif
            dest += stringNObj(dest, &obj->lambda->params, len);
            dest += sprintf(dest, " -> ");
            dest += stringNObj(dest, &obj->lambda->body, len);
            dest += sprintf(dest, ">");
#else
            dest += stringf(dest, len, "\\x%d", obj->number);
#endif
            break;
        }
        case TYPE_OTHER:
            dest += stringf(dest, len, "%p", obj->other->data);
            break;
    }

    if (!isValidType(*obj)) {
        dest += stringf(dest, len, "BAD_TYPE(%d) X%d", obj->type, obj->number);
    }

    return dest - initial;
}

int stringObj(char* dest, const Object* obj)
{
    return stringNObj(dest, obj, RESULT_LENGTH);
}

/**
 * Prints a given Object only if the DEBUG flag is set
 * @param obj The Object to debug
 */
void debugObj(const Object* obj)
{
#ifdef DEBUG
    printObj(obj);
#endif
}

#if defined(DEBUG) || defined(STANDALONE)

#define SIMPLE_TYPE(_type) case _type:\
    printf(#_type);\
    return

void printType(const Object* obj)
{
    if (!obj) {
        printf("NULL OBJECT");
        return;
    }
    switch (obj->type) {
        SIMPLE_TYPE(TYPE_NUMBER);
        SIMPLE_TYPE(TYPE_STRUCT);
        SIMPLE_TYPE(TYPE_BOOL);
        SIMPLE_TYPE(TYPE_LIST);
        SIMPLE_TYPE(TYPE_SLIST);
        SIMPLE_TYPE(TYPE_FUNC);
        SIMPLE_TYPE(TYPE_SYMBOL);
        SIMPLE_TYPE(TYPE_STRING);
        SIMPLE_TYPE(TYPE_OTHER);
        SIMPLE_TYPE(TYPE_ERROR);
        case TYPE_LAMBDA:
            printf("TYPE_LAMBDA Params:\n");
            return;
    }

    if (!isValidType(*obj)) {
        printf("UNKNOWN TYPE (%d)", obj->type);
    }
}

void _printObj(const Object* obj, int newline)
{
#ifdef DEBUG
    printType(obj);
#endif

    if (!obj) {
        printf(newline ? "\n" : "");
        return;
    }
    char temp[1024] = "";
    stringNObj(temp, obj, 1024);
    if (newline) {
        printf("%s\n", temp);
        if (obj->type == TYPE_ERROR) {
            if (obj->error && obj->error->plContext) {
                printf("%s\n", obj->error->plContext->text);
                return;
            }
        }
    } else {
        printf("%s", temp);
    }
}

/**
 * Prints the given Object and a newline.
 * Uses stringObj() to create the printed string
 * If the DEBUG flag is set, the Object's type will be printed first
 * @param obj The Object to print
 */
inline void printObj(const Object* obj)
{
    _printObj(obj, 1);
}

#endif

/**
 * Performs appropriate free() on a given Object and NULLs its ->forward
 *
 * Strings: The Object's ->string is freed
 * Lists: deleteList() is called on the Object (may recurse)
 * Lambdas: The body and params are cleaned, and the lambda itself is freed
 *
 * @param target The object to clean
 */
void cleanObject(Object* target)
{
    switch (target->type) {
        case TYPE_STRING:
        case TYPE_SYMBOL:
            if (!(target->string[-1] -= 1)) {
                free(target->string - 1);
            }
            break;
        case TYPE_LIST:
        case TYPE_SLIST:
            deleteList(target);
            break;
        case TYPE_LAMBDA:
            target->lambda->refs -= 1;
            if (!target->lambda->refs) {
                cleanObject(&target->lambda->params);
                cleanObject(&target->lambda->body);
                free(target->lambda);
            }
            break;
        case TYPE_STRUCT:
            for (int i = 0; i < getStructAt(target->structObject->definition)->fieldCount; i++) {
                cleanObject(&target->structObject->fields[i]);
            }
            free(target->structObject->fields);
            free(target->structObject);
            break;
        case TYPE_ERROR:
#ifndef SIMPLE_ERRORS
            free(target->error->plContext);
            free(target->error->context);
            free(target->error);
#endif
            break;
        case TYPE_OTHER:
            if (target->other->cleanup) {
                target->other->cleanup(target);
            }
            break;
        case TYPE_BOOL:
        case TYPE_NUMBER:
        case TYPE_FUNC:
            break;
    }
}

/**
 * Print the given object with a newline, then clean it
 * @param target The object to print and clean
 */
#ifdef STANDALONE

void printAndClean(Object* target)
{
    printObj(target);
    cleanObject(target);
}

#endif

/**
 * Frees all objects in a given list
 * Runs cleanObject() on every item in the list (may recurse)
 *
 * @param dest The list object to clean up
 */
void deleteList(Object* dest)
{
    if (!isListy(*dest)) {
        printf("Tried to delete something other than a list\n");
        return;
    }

    Object* march = dest->list;
    while (march != NULL) {
        Object* prevMarch = march;
        march = march->forward;
        cleanObject(prevMarch);
        free(prevMarch);
    }
}

void _copyList(Object* dest, const Object* src)
{
    FOR_POINTER_IN_LIST(src) {
        if (isListy(*POINTER)) {
            nf_addToList(dest, *POINTER);
            tail(dest)->list = NULL;
            _copyList(tail(dest), POINTER);
        } else if (isStringy(*POINTER)) {
            nf_addToList(dest, cloneString(*POINTER));
        } else {
            nf_addToList(dest, cloneObject(*POINTER));
        }
    }
}

/**
 * Does a deep copy of all items from `src` to `dest`
 * Does nothing if either param is NULL, or not a list type
 * Does a shallow delete of items from `dest` before copying
 * May recurse into lists in the list
 *
 * @param dest The list to copy to
 * @param src The list to copy from
 */
Object copyList(const Object* src)
{
    Object list = *src;
    list.forward = NULL;
    list.list = NULL;
    _copyList(&list, src);
    return list;
}

/**
 * Does a deep copy of all items from `src` to the end of `dest`
 * Does nothing if either param is NULL, or not a list type
 * May recurse into lists in the list
 *
 * @param dest The list to copy to
 * @param src The list to copy from
 */
void appendList(Object* dest, const Object* src)
{
    _copyList(dest, src);
}

/**
 * Returns a basic Object with NULL forward and the given type
 * @param type The type of Object to create
 * @return The created Object
 */
inline Object newObject(Type type)
{
    Object no;
    no.forward = NULL;
    no.type = type;
    return no;
}

// Returns an empty list Object
inline Object listObject()
{
    Object list = newObject(TYPE_LIST);
    list.list = NULL;
    return list;
}

// Returns an empty struct Object
inline Object structObject(int definition)
{
    Object structo = newObject(TYPE_STRUCT);
    structo.structObject = malloc(sizeof(struct StructObject));
    structo.structObject->definition = definition;
    structo.structObject->fields = malloc(sizeof(Object) * getStructAt(definition)->fieldCount);
    return structo;
}

// Returns a list Object starting with the given Object
inline Object startList(const Object start)
{
    Object list = listObject();
    nf_addToList(&list, start);
    return list;
}

inline int isNumber(const Object test)
{
    return test.type == TYPE_NUMBER;
}

inline int isListy(const Object test)
{
    return test.type == TYPE_LIST || test.type == TYPE_SLIST;
}

inline int isStringy(const Object test)
{
    return test.type == TYPE_STRING || test.type == TYPE_SYMBOL;
}

inline int isBool(const Object test)
{
    return test.type == TYPE_BOOL;
}

inline int isFuncy(const Object test)
{
    return test.type == TYPE_LAMBDA || test.type == TYPE_FUNC;
}

inline int isValidType(const Object test)
{
    switch (test.type) {
        case TYPE_NUMBER:
        case TYPE_BOOL:
        case TYPE_LIST:
        case TYPE_STRUCT:
        case TYPE_SLIST:
        case TYPE_FUNC:
        case TYPE_SYMBOL:
        case TYPE_LAMBDA:
        case TYPE_STRING:
        case TYPE_OTHER:
        case TYPE_ERROR:
            return 1;
    }

    return 0;
}

/**
 * Clones the given lambda with new allocations
 * Will behave unexpectedly if given something other than a lambda object!
 */
inline Object cloneLambda(const Object old)
{
    old.lambda->refs += 1;
    return old;
}

Object cloneString(Object obj)
{
    obj.string[-1] += 1;
    return obj;
}

// Returns an Object with a deep copy of the given Object
inline Object cloneOther(const Object src)
{
    return src.other->clone ? src.other->clone(src.other) : src;
}

Object cloneStruct(Object src);

inline Object cloneObject(const Object src)
{
    switch (src.type) {
        case TYPE_SLIST:
        case TYPE_LIST:
            return copyList(&src);
        case TYPE_LAMBDA:
            return cloneLambda(src);
        case TYPE_STRUCT:
            return cloneStruct(src);
        case TYPE_STRING:
        case TYPE_SYMBOL:
            return cloneString(src);
        case TYPE_ERROR:
            return errorWithContext(getErrorCode(src), src.error->context);
        case TYPE_OTHER:
            return cloneObject(src);
        case TYPE_BOOL:
        case TYPE_NUMBER:
        case TYPE_FUNC:
            return src;
    }

    return src;
}

Object cloneStruct(const Object src)
{
    Object structo = structObject(src.structObject->definition);
    struct StructObject* so = structo.structObject;
    for (int i = 0; i < getStructAt(so->definition)->fieldCount; i++) {
        so->fields[i] = cloneObject(src.structObject->fields[i]);
    }
    return structo;
}

inline Object numberObject(int num)
{
    Object o = newObject(TYPE_NUMBER);
    o.number = num;
    return o;
}

inline Object boolObject(int b)
{
    Object o = newObject(TYPE_BOOL);
    o.number = !!b;
    return o;
}

// Skips first and last chars! Assumed to be '"'
inline Object objFromSlice(const char* string, int len)
{
    return stringFromSlice(&string[1], len - 1);
}

inline Object stringFromSlice(const char* string, int len)
{
    Object o = symFromSlice(string, len);
    o.type = TYPE_STRING;
    return o;
}

inline Object symFromSlice(const char* string, int len)
{
    Object o = withLen(len, TYPE_SYMBOL);
    snprintf(o.string, len + 1, "%s", string);
    return o;
}

/// Creates a stringy object with room for a reference-count prefix and trailing null byte.
/// Thus, withLen(3, TYPE_STRING) will actually allocate a 5-byte "string".
inline Object withLen(int len, enum Type type)
{
    Object o = newObject(type);
    o.string = malloc(sizeof(char) * (len + 2));
    o.string[0] = 1;
    o.string += 1;
    return o;
}

inline Object constructLambda(const Object* params, const Object* body, struct Environment* env)
{
    if (!params || !body) {
        return errorObject(NULL_LAMBDA_LIST);
    }

    if (params->type != TYPE_LIST || body->type != TYPE_LIST) {
        return errorObject(LAMBDA_ARGS_NOT_LIST);
    }

    Object o = newObject(TYPE_LAMBDA);
    o.lambda = malloc(sizeof(struct Lambda));
    o.lambda->params = copyList(params);
    o.lambda->body = copyList(body);
    o.lambda->refs = 1;

    if (env) {
        Object* dest = &o.lambda->body;
        FOR_POINTER_IN_LIST(dest) {
            if (POINTER->type == TYPE_SYMBOL) {
                Object fetched = fetchFromEnvironment(POINTER->string, env);
                // TODO: Figure out why lambdas in particular break when doing this.
                if (!isError(fetched, DID_NOT_FIND_SYMBOL) && fetched.type != TYPE_LAMBDA) {
                    fetched.forward = POINTER->forward;
                    cleanObject(POINTER);
                    *POINTER = fetched;
                } else {
                    cleanObject(&fetched);
                }
            }
        }
    }
    return o;
}

inline int isError(const Object obj, const enum errorCode err)
{
    return obj.type == TYPE_ERROR && getErrorCode(obj) == err;
}

inline int bothAre(const enum Type type, const Object* obj1, const Object* obj2)
{
    return (obj1->type == type) && (obj2->type == type);
}

inline int areSameType(const Object* obj1, const Object* obj2)
{
    return obj1->type == obj2->type;
}

inline Object otherObject()
{
    Object o = newObject(TYPE_OTHER);
    o.other = malloc(sizeof(struct Other));
    return o;
}

inline Object errorObject(enum errorCode err)
{
    Object o = newObject(TYPE_ERROR);
#ifdef SIMPLE_ERRORS
    o.error = err;
#else
    o.error = malloc(sizeof(struct Error));
    o.error->code = err;
    o.error->context = NULL;
    o.error->plContext = NULL;
#endif

    return o;
}

inline enum errorCode getErrorCode(const Object obj)
{
#ifdef SIMPLE_ERRORS
    return obj.error;
#else
    return obj.error->code;
#endif
}

#ifndef SIMPLE_ERRORS

inline void errorAddContext(Object* o, const char* context, int lineNo, const char* fileName)
{
    o->error->context = calloc(sizeof(char), RESULT_LENGTH);
    char* cursor = o->error->context;
    cursor += sprintf(cursor, "%s", context);
    if (fileName) {
        sprintf(cursor, " [33m%s:%d[0m", fileName, lineNo);
    }
}

inline Object errorWithContextLineNo(enum errorCode code, const char* context, int lineNo, const char* fileName)
{
    Object o = errorObject(code);
    if (context) {
        errorAddContext(&o, context, lineNo, fileName);
    }
    return o;
}

#endif

struct Error noError()
{
    struct Error err;
    err.context = NULL;
    return err;
}
