#include <stdio.h>
#include <string.h>

#include "plfunc.h"

Object typeCheck(const char* funcName, Object* params, int length, int (*typeChecks[])(Object), int typeLength, int* failed)
{
    *failed = 1;
    if ((typeLength - 1) > length ) {
        return errorObject(NOT_ENOUGH_ARGUMENTS);
    }
    for (int i = 0; i < typeLength - 1; i++) {
        if (typeChecks[i] && !typeChecks[i](params[i])) { // TODO: Use pl func name instead of C function name.
            return /*errorObject(BAD_TYPE); */ errorWithContextLineNo(BAD_PARAMS_ON, funcName, 0, NULL);
        }
    }
    *failed = 0;
    return numberObject(0);
}

#ifndef DISABLE_TYPE_CHECKS
#define checkTypes(FUNC) int FAILED; Object ERROR = typeCheck(#FUNC, params, length, FUNC ## TypeChecks, array_length(FUNC ## TypeChecks), &FAILED); \
if (FAILED) { \
    return ERROR; \
}
#else
#define checkTypes(FUNC) ;
#endif

Object reduce(Object* params, int length, struct Environment* env)
{
    checkTypes(reduce)
    Object list = params[0];
    const Object func = params[1];
    Object total = params[2];

    if (list.type != TYPE_LIST) {
        return simpleFuncEval(func, total, list, env);
    }

    FOR_POINTER_IN_LIST(&list) {
        total = simpleFuncEval(func, total, *POINTER, env);
    }

    return total;
}

Object charAt(Object* params, int length, struct Environment* env)
{
    checkTypes(charAt)
    Object string = params[0];
    Object at = params[1];

    Object c = withLen(1, string.type);
    c.string[1] = '\0';

    for (int i = 0; i < at.number; i++) {
        if (string.string[i] == '\0') {
            c.string[0] = '\0';
            return c;
        }
    }

    c.string[0] = string.string[at.number];
    return c;
}

Object filter(Object* params, int length, struct Environment* env)
{
    checkTypes(filter)
    Object condition = params[0];
    Object list = params[1];

    Object filtered = listObject();
    FOR_POINTER_IN_LIST(&list) {
        Object cloned = cloneObject(condition);
        Object first_eval = eval(&cloned, env);
        Object testee = cloneObject(*POINTER);
        Object e = listEvalLambda(&first_eval, &testee, env);
        if (e.number) {
            nf_addToList(&filtered, testee); // May need to re-clone testee?
        }
        cleanObject(&e);
        cleanObject(&cloned);
    }
    return filtered;
}

Object append(Object* params, int length, struct Environment* env)
{
    checkTypes(append)
    Object list = params[0];
    Object newElement = params[1];

    Object newList = cloneObject(list);
    nf_addToList(&newList, cloneObject(newElement));
    return newList;
}

Object prepend(Object* params, int length, struct Environment* env)
{
    checkTypes(prepend)
    Object list = params[0];
    Object newElement = params[1];

    Object newList = listObject();
    nf_addToList(&newList, cloneObject(newElement));
    appendList(&newList, &list);
    return newList;
}

Object at(Object* params, int length, struct Environment* env)
{
    checkTypes(at)
    Object index = params[0];
    Object list = params[1];

    const Object* found = itemAt(&list, index.number);
    if (found) {
        return cloneObject(*found);
    } else {
        return errorObject(INDEX_PAST_END);
    }
}

Object rest(Object* params, int length, struct Environment* env)
{
    checkTypes(rest)
    Object list = params[0];
    if (!isListy(list)) {
        return errorObject(NOT_A_LIST);
    }

    Object ret = listObject();
    Object* l = &list;
    FOR_POINTER_IN_LIST(l) {
        if (POINTER == l->list) {
            continue;
        }
        nf_addToList(&ret, cloneObject(*POINTER));
    }
    return ret;
}

Object reverse(Object* params, int length, struct Environment* ignore2)
{
    checkTypes(reverse)
    Object _list = params[0];

    const Object* list = &_list;
    Object rev = listObject();

    Object* tail = NULL;
    FOR_POINTER_IN_LIST(list) {
        Object* oldTail = tail;
        allocObject(&tail, cloneObject(*POINTER));
        if (oldTail) {
            tail->forward = oldTail;
        }
    }

    rev.list = tail;
    return rev;
}

Object isNum(Object* params, int length, struct Environment* env)
{
    checkTypes(isNum)
    Object test = params[0];

    return test.type == TYPE_NUMBER ? boolObject(1) : boolObject(0);
}

Object isList(Object* params, int length, struct Environment* env)
{
    checkTypes(isList)
    Object test = params[0];

    return test.type == TYPE_LIST ? boolObject(1) : boolObject(0);
}

Object isString(Object* params, int length, struct Environment* env)
{
    checkTypes(isString)
    Object test = params[0];

    return test.type == TYPE_STRING ? boolObject(1) : boolObject(0);
}

Object charVal(Object* params, int length, struct Environment* env)
{
    checkTypes(charVal)
    Object test = params[0];

    return numberObject(test.string[0]);
}

Object isErr(Object* params, int length, struct Environment* env)
{
    checkTypes(isErr)
    Object test = params[0];

    return test.type == TYPE_ERROR ? boolObject(1) : boolObject(0);
}

Object parseEvalO(Object* params, int length, struct Environment* env)
{
    Object text = params[0];

    switch (text.type) {
        case TYPE_SYMBOL: {
            Object string = eval(&text, env);
            Object parsed = parseEval(string.string, env);
            cleanObject(&string);
            return parsed;
        }
        case TYPE_SLIST:
            return evalList(&text, env);
        case TYPE_STRING:
            return parseEval(text.string, env);
        default:
            return errorObject(CAN_ONLY_EVAL_STRINGS);
    }
}

#ifdef STANDALONE
#define CAT_MAX 1024
#else
#define CAT_MAX 64
#endif

Object _catObjects(Object obj1, Object obj2, struct Environment* env)
{
    Object evalObj1 = eval(&obj1, env);
    Object evalObj2 = eval(&obj2, env);

    char str1[CAT_MAX] = "";
    char str2[CAT_MAX] = "";
    int length = stringObj(str1, &evalObj1);
    length += stringObj(str2, &evalObj2);

    cleanObject(&evalObj1);
    cleanObject(&evalObj2);

    Object o = withLen(length, TYPE_STRING);
    sprintf(o.string, "%s%s", str1, str2);

    return o;
}

Object catObjects(Object* params, int length, struct Environment* env)
{
    checkTypes(catObjects)

    Object string = stringFromSlice("", 0);
    if (length == 0) {
        return string;
    }
    for (int i = 0; i < length; i++) {
        Object newString = _catObjects(string, params[i], env);
        cleanObject(&string);
        string = newString;
    }
    return string;
}


Object len(Object* params, int length, struct Environment* env)
{
    checkTypes(len)

    return numberObject(listLength(&params[0]));
}

#define BASIC_MATH(_name, _op) \
Object _name(Object* params, int length, struct Environment* env) \
{ \
    if (length == 0) { \
        return numberObject(0); \
    } \
    Object sum = numberObject(params[0].number); \
    for (int i = 1; i < length; i++) { \
        sum.number _op params[i].number; \
    } \
    return sum; \
}

BASIC_MATH(add, +=)

BASIC_MATH(sub, -=)

BASIC_MATH(mul, *=)

BASIC_MATH(dvi, /=)

BASIC_MATH(mod, %=)

int areEqual(const Object* obj1, const Object* obj2);

int listEquality(const Object* list1, const Object* list2)
{
    Object* element1, *element2;
    for (element1 = (list1)->list, element2 = (list2)->list;
        element1 != ((void*) 0) && element2 != ((void*) 0);
        element1 = element1->forward, element2 = element2->forward) {
        if (!areEqual(element1, element2)) {
            return 0;
        }
    }
    return (element1 == NULL && element2 == NULL);
}

int areEqual(const Object* obj1, const Object* obj2)
{
    const int n1 = obj1->number;
    const int n2 = obj2->number;

    if (bothAre(TYPE_STRING, obj1, obj2)) {
        return !strcmp(obj1->string, obj2->string);
    }
    if (bothAre(TYPE_LIST, obj1, obj2)) {
        return listEquality(obj1, obj2);
    }
    return n1 == n2 && areSameType(obj1, obj2);
}

Object equ(Object* params, int length, struct Environment* env)
{
    if (length < 2) {
        return errorObject(NOT_ENOUGH_ARGUMENTS);
    }
    int bool = 1;
    for (int i = 0; i < length - 1; i++) {
        if (!areEqual(&params[i], &params[i + 1])) {
            bool = 0;
            break;
        }
    }
    return boolObject(bool);
}

Object or(Object* params, int length, struct Environment* env)
{
    if (length < 2) {
        return errorObject(NOT_ENOUGH_ARGUMENTS);
    }
    int bool = 0;
    for (int i = 0; i < length - 1; i++) {
        if (params[i].number || params[i + 1].number) {
            bool = 1;
            break;
        }
    }
    return boolObject(bool);
}

#define BASIC_COMPARISON(NAME, OP)\
Object NAME(Object* params, int length, struct Environment* env) \
{ \
    if (length < 2) { \
        return errorObject(NOT_ENOUGH_ARGUMENTS); \
    } \
    int bool = 1; \
    for (int i = 0; i < length - 1; i++) {                       \
        if (!(params[i].number OP params[i + 1].number)) { \
            bool = 0; \
            break; \
        } \
    } \
    return boolObject(bool); \
}

BASIC_COMPARISON(greaterThan, >)

BASIC_COMPARISON(lessThan, <)

BASIC_COMPARISON(and, &&)

#ifdef STANDALONE

Object print(Object* params, int length, struct Environment* env)
{
    for (int i = 0; i < length; i++) {
        _printObj(&params[i], 0);
    }
    return numberObject(0);
}

Object numToChar(Object* params, int length, struct Environment* env)
{
    checkTypes(numToChar)
    Object c = params[0];

    if (c.type != TYPE_NUMBER) {
        return errorObject(BAD_NUMBER);
    }
    char ch[1] = {c.number};
    return stringFromSlice(ch, 1);
}

Object printEnvO(Object* params, int length, struct Environment* env)
{
    printEnv(global());
    return numberObject(0);
}

Object takeInput(Object* params, int length, struct Environment* env)
{
    Object prompt = params[0];
    if (length > 0 && prompt.type == TYPE_STRING) {
        printf("%s", prompt.string);
    }
    char input[256] = "";
    if (fgets(input, 256, stdin)) {
        return stringFromSlice(input, strlen(input) - 1);
    }
    return errorWithContext(NULL_PARSE, "fgets() error");
}

Object loadFile(Object* params, int length, struct Environment* env)
{
    checkTypes(loadFile)
    Object filename = params[0];

    if (isStringy(filename)) {
        readFile(filename.string, env);
        return numberObject(0);
    }
    return numberObject(1);
}

Object systemCall(Object* params, int length, struct Environment* env)
{
    checkTypes(systemCall)
    Object process = params[0];

    if (isStringy(process)) {
        return numberObject(system(process.string));
    }
    return numberObject(255);
}

Object help(Object* params, int length, struct Environment* env)
{
    Object symbol = params[0];

    char* help = getHelp(symbol.string);
    Object helpText = newObject(TYPE_STRING);
    helpText.string = help;
    return helpText;
}

char* readFileToString(FILE* input)
{
    size_t capacity = 128;
    char* string = malloc(sizeof(char) * capacity);
    int c;
    int i = 1; // Skip refCount

    string[0] = 1; // Set refCount
    while ((c = fgetc(input)) != EOF) {
        string[i] = c;
        i++;
        if (i == capacity) {
            char* prev = string;
            capacity *= 2;
            string = malloc(sizeof(char) * capacity);
            string += 1;
            memcpy(string, prev, sizeof(char) * (capacity / 2));
            free(prev);
        }
    }
    string[i] = '\0';

    return string + 1; // Offset past refCount
}

Object readFileToObject(Object* params, int length, struct Environment* env)
{
    checkTypes(readFileToObject)
    Object filename = params[0];

    if (filename.type != TYPE_STRING) {
        return errorObject(NULL_PARSE);
    }

    FILE* file = fopen(filename.string, "r");
    if (!file) {
        return errorObject(NULL_PARSE);
    }

    Object string = newObject(TYPE_STRING);
    string.string = readFileToString(file);
    return string;
}

#endif // STANDALONE
